var g_ball_radius = 3.0;
var g_field_center_x = 320.0;
var g_field_center_z = 320.0;
var g_field_length = 450.0 - g_ball_radius * 2.0;
var g_field_width = 281.25 - g_ball_radius * 2.0;
var g_goal_width = 96.0 - g_ball_radius * 2.0;
var g_field_min_x = g_field_center_x - g_field_length * 0.5;
var g_field_max_x = g_field_center_x + g_field_length * 0.5;
var g_field_min_z = g_field_center_z - g_field_width * 0.5;
var g_field_max_z = g_field_center_z + g_field_width * 0.5;

var g_ball_position = [ 0.0, 0.0, 0.0 ];
var g_ball_velocity = [ 0.0, 0.0, 0.0 ];
var g_ball_rotation = [
	1.0, 0.0, 0.0,
	0.0, 1.0, 0.0,
	0.0, 0.0, 1.0
];
var g_ball_decel = 5.0;

var g_no_score = { time:0, side:0, check:false };
var g_current_throw = { n:0, t:0, p: [ g_field_center_x, g_ball_radius, g_field_center_z ], v: [ 0.0, 0.0, 0.0 ], stop_time:0, score_info:g_no_score };
var g_throw_queue = [];

var g_ball_index = mx.add_statue(g_current_throw.p[0], g_current_throw.p[1], g_current_throw.p[2], 0.0,
 "@motoball_media/ball.jm", "@motoball_media/ball.png", "null");

var g_player_slot = mx.get_player_slot();
var g_player_radius = 3.0;

var g_threw_ball = false;

var g_red_score = 0;
var g_blue_score = 0;

var g_bounce_sound = mx.add_sound("@motoball_media/bounce.sw")
mx.set_sound_freq(g_bounce_sound, 22050);
mx.set_sound_vol(g_bounce_sound, 1.0);

var g_goal_sound = mx.add_sound("@motoball_media/goal.sw")
mx.set_sound_freq(g_goal_sound, 22050);
mx.set_sound_vol(g_goal_sound, 100.0);

var g_teams = [];
var g_team_badges = [];
var g_blue_side = -1;
var g_red_side = 1;
var g_teams_set = false;

var g_goal_limit = mx.get_finish_laps();
var g_time_limit = mx.get_finish_time();

function assign_teams(seconds) {
	var i, n, p, s, blue_count, red_count;

	if (g_teams_set)
		return;

	if (seconds < 5.0)
		return;

	g_teams_set = true;

	blue_count = 0;
	red_count = 0;

	n = mx.get_running_count();

	for (i = 0; i < n; i++) {
		s = mx.get_running_order_slot(i);
		p = mx.get_position(s);
		if (p[0] < g_field_center_x) {
			g_teams[s] = g_blue_side;
			if (blue_count < 20)
				g_team_badges[s] = blue_count++;
		} else {
			g_teams[s] = g_red_side;
			if (red_count < 20)
				g_team_badges[s] = 20 + red_count++;
		}
	}
}

function move_team_markers() {
	var p, s;

	for (s in g_teams) {
		s = parseInt(s, 10);
		p = mx.get_position(s);
		if (s in g_team_badges)
			mx.move_billboard(g_team_badges[s], p[0], p[1] + 5.0, p[2]);
	}
}

function mirror_modulo(r, f, origin, range) {
	var i;

	f = (f - origin) / range;
	i = Math.floor(f);
	f = f - i;

	if (i & 1) {
		r.mirrored = true;
		f = 1.0 - f;
	} else
		r.mirrored = false;

	r.value = origin + f * range;
}

function get_stop_time() {
	var s, v = [ 0.0, 0.0, 0.0 ];

	movv3(v, g_current_throw.v);
	return Math.sqrt(dotv3(v, v)) / g_ball_decel;
}

function in_goal(z) {
	var l, h;

//	mx.message("z: " + z);
	z = z % g_field_width;
	if (z < 0)
		z += g_field_width;

	l = g_field_width * 0.5 - g_goal_width * 0.5;
	h = g_field_width * 0.5 + g_goal_width * 0.5;

//	mx.message("zwrap: " + z + " l:" + l + " h:" + h);

	return z > l && z < h;
}

function get_time_for_position(p, v, a) {
	return -((v - Math.sqrt(v * v + 2.0 * a * p)) / a);
}

function get_score_info() {
	var a, d, i, p, s, side, stopx, t, v, x, z, zslope, zstep;

	t = g_current_throw.stop_time;

	p = [ 0.0, 0.0, 0.0 ];
	v = [ 0.0, 0.0, 0.0 ];
	a = [ 0.0, 0.0, 0.0 ];

	movv3(p, g_current_throw.p);
	p[0] -= g_field_min_x;
	p[2] -= g_field_min_z;

	movv3(v, g_current_throw.v);
	if (v[0] < 0.0) {
		p[0] = g_field_length - p[0];
		v[0] = -v[0];
		side = -1;
	} else
		side = 1;

	movv3(a, v);
	s = Math.sqrt(dotv3(a, a));
	if (s == 0.0)
		return g_no_score;
	a[0] *= 1.0 / s * -g_ball_decel;

	stopx = p[0] + v[0] * t + a[0] * 0.5 * t * t;

	if (stopx < g_field_length)
		return g_no_score;

	zslope = v[2] / v[0];
	zstep = zslope * g_field_length;

	x = g_field_length;
	z = p[2] + zslope * (x - p[0]);
//	mx.message("p[2]: " + p[2] + " zslope: " + zslope + " x:" + x + " p[0]:" + p[0]);

	while (x < stopx) {
		if (in_goal(z)) {
			t = get_time_for_position(x - p[0], v[0], a[0]);
			if (g_time_limit > 0.0) {
				if (g_current_throw.t + t > g_time_limit + mx.get_gate_drop_time())
					side = 0;
			} else if (g_red_score == g_goal_limit || g_blue_score == g_goal_limit)
				side = 0;
			return { time:t, side:side, check:true };
		}
		z += zstep;
		x += g_field_length;
		side = -side;
	}

	return g_no_score;
}

function get_ball_velocity(seconds) {
	var t, s, tstop, v = g_ball_velocity;

	t = seconds - g_current_throw.t;

	if (t >= g_current_throw.stop_time) {
		setv3(v, 0.0, 0.0, 0.0);
		return;
	}

	movv3(v, g_current_throw.v);
	s = Math.sqrt(dotv3(v, v));

	scalev3(v, 1.0 / s * -g_ball_decel * t);
	addv3(v, g_current_throw.v);
}

function get_ball_position(seconds) {
	var t, i, f, a, v, s, mm, p = g_ball_position;

	t = seconds - g_current_throw.t;

	if (t > g_current_throw.stop_time) {
		t = g_current_throw.stop_time;
	}

	a = [ 0.0, 0.0, 0.0 ];
	v = [ 0.0, 0.0, 0.0 ];

	movv3(a, g_current_throw.v);
	s = Math.sqrt(dotv3(a, a));
	if (s == 0.0) {
		movv3(p, g_current_throw.p);
		return;
	}
	scalev3(a, 1.0 / s * -g_ball_decel * 0.5 * t * t);

	movv3(v, g_current_throw.v);
	scalev3(v, t);

	movv3(p, g_current_throw.p);
	addv3(p, v);
	addv3(p, a);

	mm = {};

	mirror_modulo(mm, p[0], g_field_min_x, g_field_length);
	p[0] = mm.value;
	if (mm.mirrored)
		g_ball_velocity[0] *= -1.0;

	mirror_modulo(mm, p[2], g_field_min_z, g_field_width);
	p[2] = mm.value;
	if (mm.mirrored)
		g_ball_velocity[2] *= -1.0;
}

function move_ball(seconds) {
	var p;

	p = g_ball_position;

	mx.move_statue_absolute(g_ball_index, p[0], p[1], p[2], g_ball_rotation);
}

function throw_ball(seconds, p, v) {
	var n, nva, nvb;

	if (g_threw_ball)
		return;

	n = [ 0.0, 0.0, 0.0 ];

	movv3(n, g_ball_position);
	subv3(n, p);
	scalev3(n, 1.0 / Math.sqrt(dotv3(n, n)));

	nva = dotv3(n, v);
	nvb = dotv3(n, g_ball_velocity);

	if (nva <= nvb)
		return; /* ball is already moving away */

	scalev3(n, (nva - nvb) * 2.0);
	addv3(n, g_ball_velocity);

	mx.broadcast_script_message([
	 "throw",
	 (g_current_throw.n + 1).toFixed(0),
	 seconds.toFixed(8),
	 g_ball_position[0].toFixed(3), g_ball_position[1].toFixed(3), g_ball_position[2].toFixed(3),
	 n[0].toFixed(3), n[1].toFixed(3), n[2].toFixed(3)
	 ].join(" "));

	g_threw_ball = true;
}

function check_ball(seconds) {
	var p;

	if (g_player_slot < 0)
		return;

	p = mx.get_position(g_player_slot);

	p[1] = g_ball_position[1];

	if (distv3(p, g_ball_position) < g_ball_radius + g_player_radius)
		throw_ball(seconds, p, mx.get_velocity(g_player_slot));
}

function valid_next_throw(n, t) {
//	mx.message("checking n == current + 1");
	if (n == g_current_throw.n + 1)
		return true;
//	mx.message("checking n == current");
	if (n == g_current_throw.n)
		return t < g_current_throw.t;
//	mx.message("n not current");
	return false;
}

function apply_score(c, n) {
	if (c.score_info.side == 0)
		return;
	if (n.n == c.n)
		return;
	if (n.t > c.t + c.score_info.time) {
		if (c.score_info.side < 0) {
			g_red_score++;
//			mx.message("red score committed");
		} else {
			g_blue_score++;
//			mx.message("blue score committed");
		}
	}
}

function check_next_throw(seconds) {
	var q;

	if (g_throw_queue.length == 0)
		return;

	q = g_throw_queue[0];

	if (q.t > seconds)
		return;

	g_throw_queue.shift();

	if (valid_next_throw(q.n, q.t)) {
		apply_score(g_current_throw, q);
		g_current_throw = q;
		g_current_throw.stop_time = get_stop_time();
		g_current_throw.score_info = get_score_info();
//		mx.message("stop: " + g_current_throw.stop_time.toFixed(3));
//		if (g_current_throw.score_info.side != 0)
//			mx.message("score: " + g_current_throw.score_info.time.toFixed(3) + " side: " + g_current_throw.score_info.side.toFixed(3));
		g_threw_ball = false;
		mx.set_sound_pos(g_bounce_sound, g_current_throw.p[0], g_current_throw.p[1], g_current_throw.p[2]);
		mx.set_sound_vel(g_bounce_sound, 0, 0, 0);
		mx.start_sound(g_bounce_sound);
	}
}

function rotation_matrix(u, a) {
	var x, y, z, c, s, omc;

	x = u[0];
	y = u[1];
	z = u[2];
	s = Math.sin(a);
	c = Math.cos(a);
	omc = 1.0 - c;

	return [
		c + x * x * omc,
		y * x * omc + z * s,
		z * x * omc - y * s,

		x * y * omc - z * s,
		c + y * y * omc,
		z * y * omc + x * s,

		x * z * omc + y * s,
		y * z * omc - x * s,
		c + z * z * omc
	];
}

function roll_ball(seconds) {
	var r, u, a, v, s;
	var step = 1.0 / 128.0;

	v = [];
	movv3(v, g_ball_velocity);
	v[1] = 0.0;

	s = Math.sqrt(dotv3(v, v));

	if (s < 1.0 / 1024.0)
		return;

	scalev3(v, 1.0 / s);

	u = [];
	crossv3(u, [ 0.0, 1.0, 0.0 ], v);

	/* angle = length of velocity vector / ball circumference */
	a = s / g_ball_radius;

	r = rotation_matrix(u, a * step);

	mmul3x3(g_ball_rotation, r);
	remove_scale_and_shear(g_ball_rotation);
}

function red_score() {
	if (g_current_throw.score_info.check == false && g_current_throw.score_info.side < 0)
		return g_red_score + 1;
	return g_red_score;
}

function blue_score() {
	if (g_current_throw.score_info.check == false && g_current_throw.score_info.side > 0)
		return g_blue_score + 1;
	return g_blue_score;
}

function check_score(seconds) {
	if (g_current_throw.score_info.check == true && seconds >= g_current_throw.t + g_current_throw.score_info.time) {
		g_current_throw.score_info.check = false;
		setv3(g_current_throw.p, g_field_center_x, g_ball_radius, g_field_center_z);
		setv3(g_current_throw.v, 0.0, 0.0, 0.0);
		g_current_throw.stop_time = 0;

		mx.set_sound_pos(g_goal_sound, g_field_center_x, g_ball_radius, g_field_center_z);
		mx.set_sound_vel(g_goal_sound, 0.0, 0.0, 0.0);
		mx.start_sound(g_goal_sound);

		mx.message("Goal! Red:" + red_score().toFixed(0) + " Blue:" + blue_score().toFixed(0));
	}
}

function handle_script_message(msg) {
	var a, i, n, t, x, y, z, xv, yv, zv;

//	mx.message("r: " + msg);

	a = msg.split(" ");

	if (a.length == 0)
		return;

	if (a[1] == "throw") {
		i = 2;
		n = parseInt(a[i++], 10);
		t = parseFloat(a[i++]);
		x = parseFloat(a[i++]);
		y = parseFloat(a[i++]);
		z = parseFloat(a[i++]);
		xv = parseFloat(a[i++]);
		yv = parseFloat(a[i++]);
		zv = parseFloat(a[i++]);

		g_throw_queue.push({ n:n, t:t, p:[ x, y, z ], v:[ xv, yv, zv ] });
	}
}


mx.script_message_handler = handle_script_message;

function game_over() {
	if (g_time_limit > 0.0)
		return mx.seconds > g_time_limit + mx.get_gate_drop_time();
	else
		return Math.max(red_score(), blue_score()) >= g_goal_limit;
}

function time_string(s) {
	var m;

	m = Math.floor(s / 60.0);
	s = Math.floor(s - m * 60.0);
	return m.toFixed(0) + ":" + (s < 10.0 ? "0" + s.toFixed(0) : s.toFixed(0));
}

var g_current_scoreboard = "";

function update_scoreboard() {
	var s, r, b, t, head, red, blue;

	r = red_score();
	b = blue_score();

	if (game_over()) {
		if (r == b)
			head = "TIE GAME";
		else if (r > b)
			head = "RED WINS";
		else
			head = "BLUE WINS";
	} else if (g_time_limit > 0.0) {
		t = mx.get_gate_drop_time();
		t = (t < 0.0) ? g_time_limit : (g_time_limit + t) - mx.seconds;
		head = time_string(t);
	} else
		head = g_goal_limit.toFixed(0) + " GOALS";

	red = red_score().toFixed(0);
	blue = blue_score().toFixed(0);

	s = head + red + blue;
	if (s == g_current_scoreboard)
		return;
	g_current_scoreboard = s;

	mx.begin_custom_frame(g_screen_tid);
	mx.paste_custom_frame(g_screen_tid, 0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0);
	draw_text_centered(g_screen_font, 0.5, 0.125, head);
	draw_text_centered(g_screen_font, 0.34375, 0.5, red);
	draw_text_centered(g_screen_font, 0.65625, 0.5, blue);
	mx.end_custom_frame(g_screen_tid);
}

function handle_frame(seconds) {
	g_frame_handler_prev(seconds);

	try {
		move_team_markers();
		move_ball(seconds);
		update_scoreboard();
	} catch (e) {
		mx.message("error: " + e.toString());
	}
}

var g_frame_handler_prev = mx.frame_handler;
mx.frame_handler = handle_frame;

function handle_tic(seconds) {
	g_tic_handler_prev(seconds);

	try {
		assign_teams(seconds);
		check_next_throw(seconds);
		get_ball_velocity(seconds);
		get_ball_position(seconds);
		roll_ball(seconds);
		check_ball(seconds);
		check_score(seconds);
	} catch (e) {
		mx.message("error: " + e.toString());
	}
}

var g_tic_handler_prev = mx.tic_handler;
mx.tic_handler = handle_tic;

/* 3d vector functions */
function setv3(r, x, y, z) { r[0] = x; r[1] = y; r[2] = z; }
function movv3(r, a) { r[0] = a[0]; r[1] = a[1]; r[2] = a[2]; }
function addv3(r, a) { r[0] += a[0]; r[1] += a[1]; r[2] += a[2]; }
function subv3(r, a) { r[0] -= a[0]; r[1] -= a[1]; r[2] -= a[2]; }
function scalev3(r, a) { r[0] *= a; r[1] *= a; r[2] *= a; }
function dotv3(a, b) { return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]; }
function normv3(r) { scalev3(r, 1.0 / Math.sqrt(dotv3(r, r))); }
function distv3(a, b) { var t = [ 0, 0, 0 ]; movv3(t, a); subv3(t, b); return Math.sqrt(dotv3(t, t)); }

/* [ 0, 1, 0 ] x [ 0, 0, 1] == [ 1, 0, 0 ] */
function crossv3(r, a, b) {
	var x, y, z;
	x = a[1] * b[2] - a[2] * b[1];
	y = a[2] * b[0] - a[0] * b[2];
	z = a[0] * b[1] - a[1] * b[0];
	setv3(r, x, y, z);
}

function xformv3(r, m) {
	var x, y, z;

	x = r[0];
	y = r[1];
	z = r[2];

	r[0] = m[0] * x + m[3] * y + m[6] * z;
	r[1] = m[1] * x + m[4] * y + m[7] * z;
	r[2] = m[2] * x + m[5] * y + m[8] * z;
}

function mmul3x3(r, m) {
	var v = [];

	v[0] = r[0];
	v[1] = r[1];
	v[2] = r[2];

	xformv3(v, m);

	r[0] = v[0];
	r[1] = v[1];
	r[2] = v[2];

	v[0] = r[3];
	v[1] = r[4];
	v[2] = r[5];

	xformv3(v, m);

	r[3] = v[0];
	r[4] = v[1];
	r[5] = v[2];


	v[0] = r[6];
	v[1] = r[7];
	v[2] = r[8];

	xformv3(v, m);

	r[6] = v[0];
	r[7] = v[1];
	r[8] = v[2];
}

function remove_scale_and_shear(r) {
	var x, y, z, i;

	x = r.slice(0, 3);
	y = r.slice(3, 6);
	z = r.slice(6, 9);

	normv3(z);
	crossv3(x, y, z);
	normv3(x);
	crossv3(y, z, x);

	for (i = 0; i < 3; i++) {
		r[i] = x[i];
		r[i + 3] = y[i];
		r[i + 6] = z[i];
	}
}

/*
################################################################################
## Video screen display
##
################################################################################
*/
var g_screen_tid = mx.read_texture("@motoball_media/screen.seq");

var g_screen_font = {
tid: g_screen_tid,
xoffset: 0,
yoffset: 0,
xscale: 1,
yscale: 1,
overlap: 8,
width: 1024,
height: 512,
line_height: 64,
coords: [ [ 0, 0, 24 ], [ 24, 0, 32 ], [ 56, 0, 48 ], [ 104, 0, 40 ],
[ 144, 0, 56 ], [ 200, 0, 48 ], [ 248, 0, 16 ], [ 264, 0, 32 ],
[ 296, 0, 32 ], [ 328, 0, 32 ], [ 360, 0, 40 ], [ 400, 0, 24 ],
[ 424, 0, 32 ], [ 456, 0, 24 ], [ 480, 0, 32 ], [ 512, 0, 40 ],
[ 552, 0, 40 ], [ 592, 0, 48 ], [ 640, 0, 48 ], [ 688, 0, 48 ],
[ 736, 0, 40 ], [ 776, 0, 40 ], [ 816, 0, 40 ], [ 856, 0, 40 ],
[ 896, 0, 40 ], [ 936, 0, 24 ], [ 960, 0, 32 ], [ 0, 64, 40 ],
[ 40, 64, 40 ], [ 80, 64, 40 ], [ 120, 64, 32 ], [ 152, 64, 56 ],
[ 208, 64, 48 ], [ 256, 64, 48 ], [ 304, 64, 48 ], [ 352, 64, 48 ],
[ 400, 64, 48 ], [ 448, 64, 48 ], [ 496, 64, 48 ], [ 544, 64, 48 ],
[ 592, 64, 32 ], [ 624, 64, 40 ], [ 664, 64, 48 ], [ 712, 64, 40 ],
[ 752, 64, 56 ], [ 808, 64, 48 ], [ 856, 64, 48 ], [ 904, 64, 48 ],
[ 952, 64, 48 ], [ 0, 128, 48 ], [ 48, 128, 48 ], [ 96, 128, 48 ],
[ 144, 128, 48 ], [ 192, 128, 48 ], [ 240, 128, 64 ], [ 304, 128, 56 ],
[ 360, 128, 48 ], [ 408, 128, 48 ], [ 456, 128, 40 ], [ 496, 128, 24 ],
[ 520, 128, 40 ], [ 560, 128, 40 ], [ 600, 128, 40 ], [ 640, 128, 24 ],
[ 664, 128, 40 ], [ 704, 128, 48 ], [ 752, 128, 40 ], [ 792, 128, 48 ],
[ 840, 128, 40 ], [ 880, 128, 40 ], [ 920, 128, 48 ], [ 968, 128, 40 ],
[ 0, 192, 32 ], [ 32, 192, 32 ], [ 64, 192, 48 ], [ 112, 192, 32 ],
[ 144, 192, 56 ], [ 200, 192, 40 ], [ 240, 192, 40 ], [ 280, 192, 48 ],
[ 328, 192, 48 ], [ 376, 192, 40 ], [ 416, 192, 40 ], [ 456, 192, 32 ],
[ 488, 192, 40 ], [ 528, 192, 40 ], [ 568, 192, 56 ], [ 624, 192, 48 ],
[ 672, 192, 48 ], [ 720, 192, 40 ], [ 760, 192, 40 ], [ 800, 192, 16 ],
[ 816, 192, 40 ], [ 856, 192, 40 ] ]
};

function make_cell_map(f, s, d) {
	var i, m;

	m = [];

	d = s.indexOf(d);

	for (i = 0; i < 128; i++)
		m[String.fromCharCode(i)] = d;

	for (i = 0; i < s.length; i++)
		m[s[i]] = i;

	f.map = m;
}

make_cell_map(g_screen_font, 
 "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~",
 "*");

var g_current_screen = "";

function get_cell_coords(f, c) {
	if (!(c in f.map))
		mx.message("Not in map " + c);
	return f.coords[f.map[c]];
}

function draw_text(f, x, y, s, draw) {
	var i, c, sx, sy, dx, dy, w, h;

	dx = x;
	dy = y;

	for (i = 0; i < s.length; i++) {
		if (s[i] == "\n") {
			dx = x;
			dy += f.line_height / f.height * f.yscale;
			continue;
		}
		if (s[i] == " ") {
			c = get_cell_coords(f, ","); /* comma about as wide as space */
			dx += c[2] / f.width * f.xscale;
			continue;
		}
		c = get_cell_coords(f, s[i]);
		sx = c[0] / f.width * f.xscale + f.xoffset;
		sy = c[1] / f.height * f.yscale + f.yoffset;
		w = c[2] / f.width * f.xscale;
		h = f.line_height / f.height * f.yscale;
		if (draw)
			mx.paste_custom_frame(f.tid, 1, sx, sy, dx, dy, w, h);
		dx += w - f.overlap / f.width * f.xscale;
	}

	return dx - x;
}

function draw_text_centered(f, x, y, s) {
	var w;

	w = draw_text(f, x, y, s, false);
	draw_text(f, x - w * 0.5, y, s, true);
}

function show_text(s) {
	if (s == g_current_screen)
		return
	g_current_screen = s;
	mx.begin_custom_frame(g_screen_tid);
	mx.paste_custom_frame(g_screen_tid, 0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0);
	draw_text(g_screen_font, 0, 0, s, true);
	mx.end_custom_frame(g_screen_tid);
}

show_text("\n**** COMMODORE 64 BASIC V2 ****\n\n64K RAM SYSTEM\n38911 BASIC BYTES FREE\n\nREADY.");
